from django.conf import settings

USER_CATEGORIES_REQUIRING_RESET = getattr(
    settings, "RESETTY_USER_CATEGORIES_REQUIRING_RESET", ["is_staff"]
)
RESET_PASSWORD_DELTA_DAYS = getattr(settings, "RESETTY_RESET_PASSWORD_DELTA_DAYS", 90)
RESET_PASSWORD_URL = getattr(settings, "RESETTY_RESET_PASSWORD_URL", None)
REDIRECT_EXCLUDED_PATHS = getattr(settings, "RESETTY_REDIRECT_EXCLUDED_PATHS", [])



from datetime import date
from django.http import HttpResponseRedirect

from .http import reset_password_redirect_url
from .models import ResetPasswordExtra
from .reset_service import (
    # password_due,
    path_excluded_from_redirect,
    should_reset_password,
)


class ResetPasswordMiddleware(object):
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if should_reset_password(request.user) and not path_excluded_from_redirect(
            request.path
        ):
            # handle ?next url so a user goes back after resetting
            return HttpResponseRedirect(
                reset_password_redirect_url(request.user, request.path)
            )

        # the code above happens before subsequent middlewares in the stack and response

        response = self.get_response(request)

        # the code below will happen after response and previous middlewares in the stack

        return response



from django.contrib.auth import get_user_model
from django.db.models.signals import post_save, pre_save
from django.dispatch import receiver

# pull today from the service so it is consistent across this package
from .reset_service import today, user_within_categories_requiring_reset
from .models import ResetPasswordExtra

User = get_user_model()


def user_should_set_password_details_skipping_reset(user):
    return not hasattr(
        user, "password_details"
    ) and not user_within_categories_requiring_reset(user)


def has_attr_present(obj, attr_name):
    return hasattr(obj, attr_name) and getattr(obj, attr_name)


@receiver(post_save, sender=User)
def create_password_details(sender, instance, created, **kwargs):
    """ Makes sure we save the password details only after the user instance is saved."""
    # test that instance has an id since save can be called without commit
    if (
        created
        and has_attr_present(instance, "id")
        and user_should_set_password_details_skipping_reset(instance)
    ):
        instance.password_details = ResetPasswordExtra(
            user=instance, password_last_updated_at=today()
        )

    if hasattr(instance, "password_details") and has_attr_present(
        instance.password_details, "user_id"
    ):
        instance.password_details.save()


@receiver(pre_save, sender=User)
def set_last_password_update(sender, **kwargs):
    user = kwargs.get("instance")

    if user:
        new_password = user.password
        old_password = find_password_from_db(user)

        if (
            new_password != old_password
            or user_should_set_password_details_skipping_reset(user)
        ):
            # this makes sure that the update only happens on password change.
            create_or_update_password_last_update(user, today())


def create_or_update_password_last_update(user, the_date):
    # assigns a new password details object or updates an existing one.
    # BUT will only be saved after the user instance is saved.
    if hasattr(user, "password_details"):
        password_details = user.password_details
        password_details.password_last_updated_at = the_date
    else:
        password_details = ResetPasswordExtra(
            user=user, password_last_updated_at=the_date
        )


def find_password_from_db(user):
    if User.objects.filter(pk=user.pk).exists():
        return User.objects.get(pk=user.pk).password
    return None



from django.conf import settings

settings.configure(
    {
        "RESETTY_USER_CATEGORIES_REQUIRING_RESET": ["is_staff"],
        "RESETTY_RESET_PASSWORD_DELTA_DAYS": 30,
    },
)

import arrow
import pytest
from pytest_mock import mocker
from collections import namedtuple

from resetty.resetty.reset_service import (
    calculate_days_passed,
    password_due,
    path_excluded_from_redirect,
    should_reset_password,
)
from resetty.resetty import app_settings

User = namedtuple(
    "User", ["name", "is_authenticated", "is_staff", "is_superuser", "password_details"]
)
User.__new__.__defaults__ = ("Ignacio", False, False, False, None)
Password = namedtuple("ResetPasswordExtra", ["password_last_updated_at"])


@pytest.fixture
def user_with_no_password_details():
    # user does not have the 'password_details' attribute
    UserWithNoPswdDetails = namedtuple("User", ["name", "is_authenticated"])
    return UserWithNoPswdDetails(name="Ignacio", is_authenticated=True)


@pytest.fixture
def user_with_password_updated_recently():
    ten_days_ago = arrow.utcnow().shift(days=-10).date()
    return User(
        name="Ignacio", password_details=Password(password_last_updated_at=ten_days_ago)
    )


@pytest.fixture
def user_with_password_updated_long_time_ago():
    thirty_days_ago = (
        arrow.utcnow().shift(days=-(app_settings.RESET_PASSWORD_DELTA_DAYS + 1)).date()
    )
    return User(
        name="Ignacio",
        is_authenticated=True,
        password_details=Password(password_last_updated_at=thirty_days_ago),
    )


@pytest.fixture
def staff_user(user_with_password_updated_long_time_ago):
    temp = user_with_password_updated_long_time_ago._asdict()
    temp["is_staff"] = True
    return User(**temp)


@pytest.fixture
def super_user(user_with_password_updated_long_time_ago):
    temp = user_with_password_updated_long_time_ago._asdict()
    temp["is_superuser"] = True
    return User(**temp)


def test_calculate_days_passed():
    now = arrow.utcnow()
    # test 31 days before
    days_before_as_arrow = now.floor("day").shift(days=-31)
    days_before_as_date = days_before_as_arrow.datetime.date()
    assert calculate_days_passed(days_before_as_date) == 31


def test_password_last_update_has_not_been_set(user_with_no_password_details):
    assert password_due(user_with_no_password_details) is True


def test_password_last_update_is_none():
    user = User(name="Ignacio", is_authenticated=True, password_details=None)
    assert password_due(user) is True


def test_user_with_last_update_within_threshold(user_with_password_updated_recently):
    assert password_due(user_with_password_updated_recently) is False


def test_user_with_last_update_outside_threshold(
    user_with_password_updated_long_time_ago,
):
    assert password_due(user_with_password_updated_long_time_ago) is True


def test_user_needing_reset_not_authenticated(
    mocker, user_with_password_updated_long_time_ago
):
    temp = user_with_password_updated_long_time_ago._asdict()
    temp["is_authenticated"] = False
    non_auth_user = User(**temp)
    assert should_reset_password(non_auth_user) is False


def test_staff_user_requires_reset_by_default(staff_user):
    assert should_reset_password(staff_user) is True


def test_staff_user_requires_reset_by_default(monkeypatch, staff_user):
    monkeypatch.setattr(app_settings, "USER_CATEGORIES_REQUIRING_RESET", [])
    assert should_reset_password(staff_user) is False


def test_super_user_skips_reset_by_default(super_user):
    assert should_reset_password(super_user) is False


def test_super_user_requires_reset_by_settings(monkeypatch, super_user):
    monkeypatch.setattr(
        app_settings, "USER_CATEGORIES_REQUIRING_RESET", ["is_superuser"]
    )
    assert should_reset_password(super_user) is True


def test_default_paths_excluded_from_redirect():
    default_excluded_paths = [
        "/accounts/reset/",
        "/accounts/reset/123/gq/"
        "/admin/login/",
        "/admin/logout/",
        "/admin/password_change/",
        "/admin/password_change/done/",
        "/reset/",
        "/reset/123/gq/"
    ]

    for path in default_excluded_paths:
        assert path_excluded_from_redirect(path)


def test_default_and_configured_paths_excluded_from_redirect(monkeypatch):
    custom_paths = ["/mypersonal/path"]
    monkeypatch.setattr(app_settings, "REDIRECT_EXCLUDED_PATHS", custom_paths)

    default_excluded_paths = [
        "/accounts/reset/",
        "/admin/login/",
        "/admin/logout/",
        "/admin/password_change/",
        "/admin/password_change/done/",
    ]

    for path in default_excluded_paths + custom_paths:
        assert path_excluded_from_redirect(path)



from django.contrib.auth import authenticate
from django.core.exceptions import ValidationError

# https://sixfeetup.com/blog/custom-password-validators-in-django


class DoNotReusePasswordValidator(object):
    def validate(self, password, user):
        if not user:
            return

        if authenticate(None, username=user.username, password=password):
            raise ValidationError(
                "Your old password can't be used", code="password_not_changed",
            )

    def get_help_text(self):
        return "Your old password can't be used"



# Generated by Django 2.2.11 on 2020-06-13 20:25

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("resetty", "0001_initial"),
    ]

    operations = [
        migrations.RemoveField(model_name="resetpasswordextra", name="id",),
        migrations.AlterField(
            model_name="resetpasswordextra",
            name="user",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                primary_key=True,
                serialize=False,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



from django.conf import settings
from django.db import models
from django.db.models.fields import DateField


class ResetPasswordExtra(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        models.CASCADE,
        primary_key=True,
        related_name="password_details",
    )
    password_last_updated_at = DateField()



import re
from datetime import datetime, timezone
from . import app_settings


def today():
    return datetime.now(timezone.utc).date()


def calculate_days_passed(last_password_update_at, target=today()):
    return (target - last_password_update_at).days


def password_has_due_date(password_details):
    last_password_update_at = password_details.password_last_updated_at
    return (
        calculate_days_passed(last_password_update_at)
        >= app_settings.RESET_PASSWORD_DELTA_DAYS
    )


def password_due(user):
    return not (
        hasattr(user, "password_details") and user.password_details
    ) or password_has_due_date(user.password_details)


def path_excluded_from_redirect(current_path):
    # add these paths to the app_settings
    excluded_paths = app_settings.REDIRECT_EXCLUDED_PATHS + [
        "/admin/login/",
        "/admin/logout/",
        "/admin/password_change/",
        "/admin/password_change/done/",
    ]

    pattern = r"\/?(.+)?\/reset\/?"

    return re.match(pattern, current_path) or any(
        [current_path == path for path in excluded_paths]
    )


def user_within_categories_requiring_reset(user):
    return any(
        [
            getattr(user, category)
            for category in app_settings.USER_CATEGORIES_REQUIRING_RESET
        ]
    )


def should_reset_password(user):
    return (
        user.is_authenticated
        and user_within_categories_requiring_reset(user)
        and password_due(user)
    )



from django.urls import reverse
from .app_settings import RESET_PASSWORD_URL


def reset_password_redirect_url(user, path):
    # default is /admin/password_change/
    redirect_to = RESET_PASSWORD_URL or reverse("admin:password_change")
    return f"{redirect_to}?next={path}"
